# 2024-6-13 ver4
# 从指定文件夹读取所有的Internet Archive（archive.org) item metadata (.json)文件，
# 将其中的.warc或.warc.gz文件生成下载链接(新增.gz .zip .rar .7z)，以下统称为warc文件
# 输入文件夹output_metadata
# 输出3个文件：
# 输出download_url_domain.txt 包含用于下载的warc的url、文件大小、从warc文件名转化后的域名，按域名字母顺序排序，用tab分隔
# 输出download_url_nodomain.txt 包含用于下载的warc的url、文件大小，按照文件大小数字排序（降序），用tab分隔
# 输出download_url_summary.txt，将从warc文件名转化后的域名分类汇总并排序,第一列是域名，第二列是该域名下所有warc的大小之和
# 
# 生成的.txt包含多个列，使用下载工具导入第一列可以直接下载，IA不限制连接数和带宽
# 
# 已知问题：metadata中可能存在warc、warc.gz、rar、7z等多种格式的文件，在没有下载并解压缩前，无法获得准确的有效文件大小，此处将压缩和不压缩的文件统一对待。
# 已知问题：本程序没有考虑metadata中同时存在.warc和.warc.gz的情形，会重复生成2个下载链接。
# 已知问题：IA对于每个item，一般有2个服务器可提供服务，本程序只采用第一个服务器（d1），如有需要，将data.get('d1')改为data.get('d2')即可生成使用第二个服务器的下载链接

import os  
import json
from collections import defaultdict
import tldextract
import re

# 设置文件夹路径  
input_dir = './output_metadata'  
output_file_w_domain = './download_url_domain.txt'
output_file_wo_domain = './download_url_nodomain.txt'
output_file_summary = './download_url_summary.txt'
  
# 创建一个列表来保存所有找到的URL  
urls = []  
  
# 遍历文件夹中的所有JSON文件  
for filename in os.listdir(input_dir):  
    if filename.endswith('.json'):  
        json_file_path = os.path.join(input_dir, filename)  
        try:  
            # 读取并解析JSON文件  
            with open(json_file_path, 'r', encoding='utf-8') as f:  
                data = json.load(f)  
            # 从metadata提取d1和dir属性的值（dl是下载服务器域名，dir是路径）
            d1 = data.get('d1')  
            dir_value = data.get('dir')  
            # 检查是否提取到了d1和dir的值  
            if not d1 or not dir_value:  
                print(f"Missing d1 or dir in {filename}")  
                continue  
            # 从meatadata获取所有的files
            files = data.get('files', [])  # 如果JSON中没有'files'键，则默认为空列表
            # 遍历所有的files，提取所需要的文件（warc、rar等）
            found_warc = False  
            for file_info in files:  
                name = file_info.get('name')  #文件名
                size = file_info.get('size')  #文件大小
                # 排除索引文件（.cdx.gz）
                #if name and (not name.endswith('.cdx.gz') or not name.endswith('.json.gz')) and (name.endswith('.warc') or name.endswith('warc.gz') or name.endswith('gz') or name.endswith('zip') or name.endswith('rar') or name.endswith('7z')):  
                if name and (name.endswith('.warc') or name.endswith('warc.gz') or name.endswith('zip') or name.endswith('rar') or name.endswith('7z')):  
                    found_warc = True  
                    # metadata中的dl、dir_value、name组合成可用的下载连接
                    url = 'https://' + d1 + '/' + dir_value + '/' + name  
                    # 尝试从warc文件名提取域名p1：从流行的顶级域名进行判断，如果文件名中包含下列顶级域名，则进入p2阶段
                    # top 100 TLDs
                    top_level_domains = [
    '.com', '.net', '.edu', '.ru', '.org', '.gov', '.dev', '.info', '.tech', '.in', '.ir', '.uk', '.au', '.de', '.ua', '.ca', '.tr', '.co', '.jp', '.vn', '.cn', '.gr',
    '.tw', '.id', '.br', '.io', '.xyz', '.it', '.nl', '.pl', '.za', '.us', '.eu', '.mx', '.ch', '.biz', '.me', '.il', '.es', '.online', '.by', '.fr', '.tk',
    '.nz', '.kr', '.cz', '.ro', '.cf', '.ar', '.club', '.my', '.tv', '.kz', '.cl', '.pk', '.pro', '.site', '.th', '.se', '.sg', '.cc',
    '.be', '.rs', '.top', '.ga', '.ma', '.hu', '.ae', '.su', '.dk', '.hk', '.at', '.ml', '.shop', '.store', '.ng', '.np', '.no', '.app', '.live',
    '.pe', '.ph', '.ie', '.lk', '.gq', '.fi', '.ai', '.sa', '.pw', '.bd', '.sk', '.ke', '.pt', '.az', '.space', '.mk', '.ge',
    '.tn', '.lt', '.to']
                    # 去掉warc文件的扩展名
                    name_trip = name.rstrip('.megawarc.warc.gz').rstrip('.warc.gz').rstrip('.warc').rstrip('.zip').rstrip('.rar').rstrip('.7z')
                    # 特殊处理__jp2
                    name_trip = name_trip.rstrip('__jp2')
                    tld_detect="=== 不存在有效的域名 ==="
                    for tld in top_level_domains:
                        if tld in name_trip:
                            # 字符串以TLD结尾
                            if name_trip.split(tld, 1)[1] == '':
                                tld_detect = name_trip.split(tld, 1)[0] + tld
                            else:
                                # 字符串不以TLD结尾，但是TLD之后的不是英文字符（为了防止.jpeg被识别为.jp域名，只有像.jpeg-20240101这种才会被计入）
                                if not name_trip.split(tld, 1)[1][0].isalpha():
                                    tld_detect = name_trip.split(tld, 1)[0] + tld
                    #处理字符串的开头，有可能文件名开头有特殊符号，例如warc/www.anonpone.com，此处从整个字符串中删掉除了.之外的所有特殊符号
                    pattern = r'[^\w\.]+'
                    tld_detect=re.sub(pattern, '', tld_detect)
                    # 尝试从warc文件名提取域名p2，使用tldextract组件
                    domain = tldextract.extract(tld_detect)
                    domain_final="=== domain not found ==="
                    if domain.suffix:
                        domain_final=domain.domain + '.' + domain.suffix
                    # 只保留1级域名，例如files.google.com=google.com
                    urls.append({'url': url, 'size': size, 'domain': domain_final, 'warcname': name})
            # 如果没有找到".warc"或"warc.gz"文件，则打印错误信息  
            if not found_warc:  
                print(f"No .warc warc.gz gz zip rar 7z files found in {filename}")  
        except json.JSONDecodeError as e:  
            print(f"Error parsing JSON file {filename}: {e}")  
        except Exception as e:  
            print(f"Error processing file {filename}: {e}")  
  
urls_w_domain = []
urls_wo_domain = []
for url_info in urls:
    # 检查 domain 是否等于 "=== domain not found ==="
    if url_info['domain'] == "=== domain not found ===":
        # 如果是，添加到 urls_wo_domain 列表
        urls_wo_domain.append(url_info)
    else:
        # 如果不是，添加到 urls_w_domain 列表
        urls_w_domain.append(url_info)

#按照识别的域名排序
# 没有找到域名的项目，按照size排序
# 找到域名的项目，按照域名排序
urls_w_domain.sort(key=lambda x: x['domain'])
urls_wo_domain.sort(key=lambda x: float(x['size']), reverse=True)

# 将所有URL、文件大小、域名写入文件
with open(output_file_w_domain, 'w', encoding='utf-8') as f_good:
    for url in urls_w_domain:  
        f_good.write(f"{url['url']}\t{url['size']}\t{url['domain']}\n")

with open(output_file_wo_domain, 'w', encoding='utf-8') as f_bad:
        for url in urls_wo_domain:  
            f_bad.write(f"{url['url']}\t{url['size']}\n")

# 按照domain将size分类汇总，写入文件
domain_sizes = defaultdict(list)
for url_info in urls:
    domain = url_info['domain']
    size = url_info['size']
    size = int(float(size)) 
    domain_sizes[domain].append(size)

size_by_domain = []
for domain, sizes in domain_sizes.items():
    total_size = sum(sizes)
    size_by_domain.append({'domain': domain, 'total_size': total_size})
# 按照总文件大小排序
size_by_domain.sort(key=lambda x: x['total_size'], reverse=True)
with open(output_file_summary, 'w', encoding='utf-8') as f:  
    for domain in size_by_domain:  
        #f.write(url + '\n')
        f.write(f"{domain['domain']}\t{domain['total_size']}\n")
